#!/usr/bin/env python3

# must use python 3 because of quick2wire interface

# AD7745 I2C address is fixed as 7'b1001000 = 0x48 
import sys
sys.path.append('/home/pi/programs/commonCode')
import time
from Q2W_i2c import *
import math
from collections import deque

class AD7745:

    # define constants
    __AD7745_address = 0x48
    __AD7745_status = 0b00000000
    __AD7745_CapData = 0b00000001
    __AD7745_VTdata = 0b00000110
    __AD7745_CapSetup = 0b00000111
    __AD7745_VTSetup = 0b00001000
    __AD7745_EXCSetup = 0b00001001
    __AD7745_Configuration = 0b00001010
    __AD7745_CapDAC_A = 0b00001011
    __AD7745_CapDAC_B = 0b00001100
    __AD7745_CapOffset = 0b00001101
    __AD7745_CapGain = 0b00001111
    __AD7745_VoltGain = 0b00010001
    
    # Constructor
    def __init__(self,Configuration=0b00010000, CapSetup=0b10100000, EXCSetup=0b00011011, debug=False):
        '''Default setup is for fully differential bridge-mode: excitation (EXA-EXB) and capacitance (CIN+ - CIN-).
        Conversion mode and timing is set by a write to Configuration Register.'''
        self.reset()
        self.write_reg(self.__AD7745_Configuration, Configuration)  # VT conversion = 20ms, Cap conversion = 20ms, Single conversion
        self.write_reg(self.__AD7745_EXCSetup, EXCSetup)  # CLKCTRL=0, EXCON=0, EXB enabled inverted, EXA enabled, Exitation level=VDD
        self.write_reg(self.__AD7745_CapSetup,CapSetup) # Enable Cap Conversion, Differential Mode
        self.debug = debug

    def reset(self, debug=False):
        Q2WwriteByte(self.__AD7745_address, 0xBF, debug=debug)  # reset the part
        time.sleep(0.0002)  # 200us required to reset the part

    def read_status(self, debug=False):
        return Q2WreadU8(self.__AD7745_address,self.__AD7745_status, debug=debug)
    
    def read_reg(self, reg, debug=False):
        data = Q2WreadU8(self.__AD7745_address,reg, debug=debug)
        return hex(data)

    def write_reg(self, reg, value, debug=False):
        Q2Wwrite8(self.__AD7745_address, reg, value, debug=debug)
        
    def writeOffset(self, offset):
        Q2Wwrite16(self.__AD7745_address, 0x0d, offset)

    def writeCapDACA(self, value, debug=False):
        self.write_reg(0x0b, value, debug=debug)

    def writeCapDACB(self, value, debug=False):
        self.write_reg(0x0c, value, debug=debug)

    def writeBothDACs(self, DACA, DACB, debug=False):
        data = (DACA << 8) + DACB
        Q2Wwrite16(self.__AD7745_address, 0x0b, data, debug=debug)

    def get_capCount(self, debug=False):
        junk, CapDataH, CapDataM, CapDataL = Q2WreadListNoReg(self.__AD7745_address,4, debug=debug)
        CapCount = (CapDataH << 16) + (CapDataM << 8) + CapDataL
        return CapCount - 0x800000
    
    def get_capacitance(self, debug=False):
        '''This should be the correct equation, but it doesn't appear to yield the correct
           result. Off by a factor of 1.5 (low). So just get the count and process that data '''
        CapCount = self.get_capCount(debug=debug)
        capacitance = float(2* 4096 * 1E-15 * 2**-24 * CapCount) # +/- 4.096pf/2**24
        return CapCount, capacitance
    
    def offsetCalibrate(self, debug=False):
        self.write_reg(self.__AD7745_Configuration, 0x25, debug=debug)
    
    def read_everything(self, debug=False):
        Status, CapDataH, CapDataM, CapDataL, VTDataH, VTDataM, VTDataL, CapSetup, VTSetup, EXCSetup, Config, CapDACA, CapDACB, CapOffsetH, CapOffsetL, CapGainH, CapGainL, VgainH, VgainL = Q2WreadListNoReg(0x48, 19, debug=debug)
        CapOffset = (CapOffsetH << 8) + CapOffsetL
        CapGain = (CapGainH << 8) + CapGainL
        CapData = (CapDataH << 16) +(CapDataM << 8) + CapDataL - 0x800000
        VTData = (VTDataH << 16) +(VTDataM << 8) + VTDataL
        Vgain = (VgainH << 8) + VgainL
        return Status, CapData, VTData, CapSetup, VTSetup, EXCSetup, Config, CapDACA, CapDACB, CapOffset, CapGain, Vgain

    def meanstdv(self,x):
        """
        Calculate mean and standard deviation of data x[]:
        mean = {\sum_i x_i \over n}
        std = math.sqrt(\sum_i (x_i - mean)^2 \over n-1)
        """
        n, mean, std = len(x), 0, 0
        for a in x:
            mean = mean + a
        mean = mean / float(n)
        for a in x:
            std = std + (a - mean)**2
        if(n > 1):
            std = math.sqrt(std / float(n-1))
        else:
            std = 0.0
        return mean, std

class rollingAverage:
    ''' Returns a simple rolling average of n most recent values. Also returns the
        average slope, in degrees F per hour, of the last n most recent rolling averages'''

    def __init__(self, nAvg=5, debug=False):
        "determine lengh of roll at instantiation"
        self.nAvg = nAvg
        self.xqueue = deque('')
        self.debug = debug
        
    def rolling_avg(self, x):
        # if the queue is empty then fill it with values of x
        if(self.xqueue == deque([])):
            for i in range(self.nAvg):
                self.xqueue.append(x)
        self.xqueue.append(x)
        self.xqueue.popleft()
        avg = 0
        for i in self.xqueue:
            avg += i
        avg = avg/float(self.nAvg)

        if self.debug:
            print("Rolling Avg:")
            for i in self.xqueue:
                print(i)
            print("avg: %f" % avg)
        return avg

class gaussianLowPass:   
    ''' performs a gaussian Low Pass filter of length n
    '''

    def __init__(self, nG=3, debug=False):
        "determine lengh of filter at instantiation"
        self.nG = nG
        self.xqueue = deque('')
        self.debug = debug
        self.sigma = nG/6
        self.weight = []
        for n in range(4*self.nG+1):
            nweight = self.normal(n, self.sigma)
            if (n == 0):
                self.weight.append(nweight)
                weight0 = nweight
            elif (nweight/weight0 > 0.05):
                self.weight.append(nweight)
                self.weight.insert(0, nweight)
        if (debug):
            print("Sigma = {0}\nRaw Weights: {1}".format(self.sigma, self.weight))
        sum = 0
        for n in range(len(self.weight)):
            sum = sum + self.weight[n]
        if debug:
            print ("Sum = {}".format(sum))
        for n in range(len(self.weight)):
            self.weight[n] = self.weight[n]/sum
        if (debug):
            print("Weighted List: {}".format(self.weight))

    def normal(self, n, sigma):
        f = 1/math.sqrt(2*math.pi)*math.exp(-((n/self.sigma)**2)/2)
        return f

    def filter(self, x):
        # if the queue is empty then fill it with values of x
        if(self.xqueue == deque([])):
            for i in range(len(self.weight)):
                self.xqueue.append(x)
        self.xqueue.append(x)
        self.xqueue.popleft()
        result = 0
        for i in range(len(self.xqueue)):
            result = result + self.xqueue[i] * self.weight[i]
        return result

class gaussianHighPass:   
    ''' performs a gaussian Hign Pass filter of length n
    The filter takes and input stream and low pass filters it with a nG length FIR filter.
    The output is the input signal, delayed by nG/2, minus the output of the low pass filter,
    which has a group delay of ~nG/2.
    '''

    def __init__(self, nG=500, debug=False):
        "determine lengh of filter at instantiation"
        self.nG = nG
        self.xqueue = deque('')
        self.debug = debug
        self.sigma = nG/6
        self.weight = []
        for n in range(4*self.nG+1):
            nweight = self.normal(n, self.sigma)
            if (n == 0):
                self.weight.append(nweight)
                weight0 = nweight
            elif (nweight/weight0 > 0.05):
                self.weight.append(nweight)
                self.weight.insert(0, nweight)
        if (debug):
            print("Sigma = {0}\nRaw Weights: {1}".format(self.sigma, self.weight))
        sum = 0
        for n in range(len(self.weight)):
            sum = sum + self.weight[n]
        if debug:
            print ("Sum = {}".format(sum))
        for n in range(len(self.weight)):
            self.weight[n] = self.weight[n]/sum
        if (debug):
            print("Weighted List: {}".format(self.weight))

    def normal(self, n, sigma):
        f = 1/math.sqrt(2*math.pi)*math.exp(-((n/self.sigma)**2)/2)
        return f

    def filter(self, x):
        # if the queue is empty then fill it with values of x
        if(self.xqueue == deque([])):
            for i in range(len(self.weight)):
                self.xqueue.append(x)
        self.xqueue.append(x)
        self.xqueue.popleft()
        LP_value = 0
        for i in range(len(self.xqueue)):
            LP_value = LP_value + self.xqueue[i] * self.weight[i]
        result = self.xqueue[len(self.xqueue)//2] - LP_value
        return result, LP_value

class HighPass:   
    ''' performs a single pole/zero IIR Hign Pass filter 
    The filter takes and input stream outputs both the LPF and HPF
    results.
    '''

    def __init__(self, Fsample = 1/16.2, Fcorner = 0.1, debug=False):
        "determine lengh of filter at instantiation"
        self.alfa = float(Fsample/(Fcorner * 2 * math.pi))
        self.debug = debug
        # define constants for the instance
        self.LPn_minus1 = 0  # low pass output Vlp * z^-1
        self.LPn = 0  # low pass output
        self.Hpn = 0  # high pass output
        self.init = 1 # initialization flag

    def filter(self, x):
        if (self.init):
            self.LPn_minus1 = x  
            self.LPn = x  
            self.Hpn = 0  
            self.init = 0
        self.HPn = x - self.LPn_minus1
        self.LPn = self.HPn/self.alfa + self.LPn_minus1
        self.LPn_minus1 = self.LPn
        return self.HPn, self.LPn
        

# test code:
if __name__=="__main__":

    cap2dig = AD7745() #initialize
    print("Initializing AD7745:")
    print("\tConversion Rate = 62ms")
    print("\tDifferential Cap Mode, EXA & EXB(inverted), Exitation level = VDD")
    
    status = cap2dig.read_status()
    print ("Status: {}".format(bin(status)))
    
    Status, CapDataH, CapDataM, CapDataL, VTDataH, VTDataM, VTDataL, CapSetup, VTSetup, EXCSetup, Config, CapDACA, CapDACB, CapOffsetH, CapOffsetL, CapGainH, CapGainL, VgainH, VgainL = cap2dig.read_everything()
    
    print("Registers after Reset:")
    print ("\tStatus: {} , should be 0x7".format(hex(Status)))
    print ("\tCapData: {0} {1} {2} , should be all zeros".format(hex(CapDataH), hex(CapDataM), hex(CapDataL)))
    print ("\tVTData: {0} {1} {2} , should be all zeros".format(hex(VTDataH), hex(VTDataM), hex(VTDataL)))
    print ("\tCapSetup: {} , should be 0xa0".format(hex(CapSetup)))
    print ("\tVTSetup: {} , should be 0x00".format(hex(VTSetup)))
    print ("\tEXCSetup: {}, should be 0x3b".format(hex(EXCSetup)))
    print ("\tConfig: {} , should be 0x22".format(hex(Config)))

    
    n=0
    conversions = 3
    print("Perform {} conversions".format(conversions))
    while (n < conversions):
        cap2dig.write_reg(0x0A, 0b00100010)  # write 0x22 into the Configuration register
        time.sleep(.1)
        capacitance = cap2dig.get_capacitance()
        print ("Capacitance = {}".format(capacitance))
        time.sleep(.001)
        n = n+1
    print("Perform offset calibration and {} more conversions:".format(3*conversions))
    n=0
    cap2dig.offsetCalibrate()
    time.sleep(0.5)
    resultArray = []
    while (n < 3 * conversions):
        cap2dig.write_reg(0x0A, 0b00100010)  # write 0x22 into the Configuration register
        time.sleep(.1)
        capacitance = cap2dig.get_capacitance()
        resultArray.append(capacitance)
        print ("Capacitance = {}".format(capacitance))
        time.sleep(.001)
        n = n+1
    
    #print("result: {}".format(resultArray))
    mean, stdev = cap2dig.meanstdv(resultArray)
    print("\nMean Offset = {}".format(mean))
    print("Std Deviation of Offset = {}\n".format(stdev))
    
    Status, CapDataH, CapDataM, CapDataL, VTDataH, VTDataM, VTDataL, CapSetup, VTSetup, EXCSetup, Config, CapDACA, CapDACB, CapOffsetH, CapOffsetL, CapGainH, CapGainL, VgainH, VgainL = cap2dig.read_everything()
    CapOffset = float((CapOffsetH << 8) + CapOffsetL - 2**15) * 10E-12/2**16
    CapGain = (CapGainH << 8) + CapGainL
    VGain = (VgainH << 8) + VgainL
    print ("Status: {}".format(hex(Status)))
    print ("CapData: {0} {1} {2}".format(hex(CapDataH), hex(CapDataM), hex(CapDataL)))
    print ("VTData: {0} {1} {2}".format(hex(VTDataH), hex(VTDataM), hex(VTDataL)))
    print ("CapSetup: {}".format(hex(CapSetup)))
    print ("VTSetup: {}".format(hex(VTSetup)))
    print ("EXCSetup: {}".format(hex(EXCSetup)))
    print ("Config: {}".format(hex(Config)))
    print ("CapDACA: {}".format(CapDACA))
    print ("CapDACB: {}".format(CapDACB))
    print ("CapOffset: {}".format(CapOffset))
    print ("CapGain: {}".format(CapGain))
    print ("Vgain: {}".format(VGain))
    
    # add 1 LSB of offset to CapDACA and measure result
    print("\nAdding 1 LSB to CapDACA")
    cap2dig.write_reg(0x0B, 0x81)  # set CapDACA to 1 LSB
    cap2dig.write_reg(0x0A, 0x22)  # set Configuration Register for single conversion
    # wait for conversion
    time.sleep(0.1)
    capacitance = cap2dig.get_capacitance()
    print("Capacitance = {}".format(capacitance))
    
    # add 4 LSB of offset to CapDACA and measure result
    print("\nAdding 4 LSB to CapDACA")
    cap2dig.write_reg(0x0B, 0x84)  # set CapDACA to 1 LSB
    cap2dig.write_reg(0x0A, 0x22)  # set Configuration Register for single conversion
    # wait for conversion
    time.sleep(0.1)
    capacitance = cap2dig.get_capacitance()
    print("Capacitance = {}".format(capacitance))

    # add 4 LSB of offset to CapDACB and measure result
    print("\nAdding 4 LSB to CapDACB")
    cap2dig.write_reg(0x0C, 0x84)  # set CapDACB to 4 LSB
    cap2dig.write_reg(0x0A, 0x22)  # set Configuration Register for single conversion
    # wait for conversion
    time.sleep(0.1)
    capacitance = cap2dig.get_capacitance()
    print("Capacitance = {}".format(capacitance))
    
    # add 5 LSB of offset to CapDACB and measure result
    print("\nAdding 4 LSB to CapDACB")
    cap2dig.write_reg(0x0C, 0x85)  # set CapDACB to 5 LSB
    cap2dig.write_reg(0x0A, 0x22)  # set Configuration Register for single conversion
    # wait for conversion
    time.sleep(0.1)
    capacitance = cap2dig.get_capacitance()
    print("Capacitance = {}".format(capacitance))
