#!/usr/bin/env python3
#-*- coding:utf-8 -*-

# must use python 3 because of quick2wire interface

'''
1. Plot 1 hour of data with 8:1 decimation
2. Plot scrolls every x seconds (determined by webpage) with new data
3. List today's date at top of page under title.
'''

from AD7745_i2c import AD7745, rollingAverage, gaussianLowPass, gaussianHighPass, HighPass
import time, json, os
from bottle import *
from datetime import datetime, timedelta, date
import RPi.GPIO as GPIO
import multiprocessing
import logging
import logging.handlers

#--logger definitions
logger = logging.getLogger(__name__) 
logger.setLevel(logging.INFO)  # Could be e.g. "TRACE", "ERROR", "" or "WARNING"
handler = logging.handlers.RotatingFileHandler("/var/log/bud_logs/seismo.log", maxBytes=10000, backupCount=4) # save 4 logs
formatter = logging.Formatter('%(asctime)s %(levelname)-8s %(message)s')
handler.setFormatter(formatter)
logger.addHandler(handler)

class MyLogger():
    '''
    A class that can be used to capture stdout and sterr to put it in the log

    '''
    def __init__(self, level, logger):
            '''Needs a logger and a logger level.'''
            self.logger = logger
            self.level = level

    def write(self, message):
        # Only log if there is a message (not just a new line)
        if message.rstrip() != "":
                self.logger.log(self.level, message.rstrip())

    def flush(self):
        pass  # do nothing -- just to handle the attribute for now

# If this script is installed as a Daemon, set this flag to True:
DAEMON = True # when run as daemon, pipe all console information to the log file
# --Replace stdout and stderr with logging to file so we can run it as a daemon
# and still see what is going on
if DAEMON :
    sys.stdout = MyLogger(logging.INFO, logger)
    sys.stderr = MyLogger(logging.ERROR, logger)

# instantiate AD7745 as cap2dig
cap2dig = AD7745(EXCSetup=0b00011011, CapSetup=0b10100000, debug=False)  # EXCA differential
# avg = rollingAverage(nAvg=4) # short length rolling average
# LPF = gaussianLowPass(nG=16, debug=False)  # short length gaussian noise filter
# HPF1 = HighPass(Fsample=16.2, Fcorner= 0.03, debug=False)  # single pole/zero high pass filter with 0.03 Hz HP corner

# global variables
writeFlag = 0
LPF1count = 0
seedArray = []
resultArray = []
gaussArray = []

# 
def getData(channel):
    '''
    This routine is called when the AD7745 conversion is complete - a GPIO4 falling edge.
    It gets the conversion result, a Cap_Count value, from the AD7745. CapCount is appended
    to the resultArray and the seedArray by the signalProcessor function. If the resultArray
    is larger than 61000 it is truncated to 60000 elements. When the seedArray equals the
    desired packet size the seedArray data is written to slist.txt and converted to
    mini-seed (mseed) format by ascii2mseed. The seedArray is emptied at this point and the
    process start over when this routine is called again.
    '''
    global resultArray, gaussArray, LPF1count, seedArray, writeFlag
    if (writeFlag):
        logger.error("SeedFile not written before next sample!")
    ts = datetime.utcnow()  # a timestamp for the seed file
    timeNow = time.time() * 1000  # this creates a unix timestamp with millisecond resolution
    CapCount = cap2dig.get_capCount()
    signalProcessor(ts,timeNow, CapCount)
    # 60000 samples is about 1 hour of data -- normal limits are 61000 max truncated to 60000.
    if (len(resultArray) > 61000):
        resultArray = resultArray[-60000:]

    packetSize = 512  # 512 is standard packet size for mseed.
    if(len(seedArray) == packetSize):
        # write data stream to ascii file
        writeFlag = 1
        #logger.info("Writing seed file.")
        asciiFile = open('/home/pi/programs/Seismo/slist.ascii', 'w')
        asciiFile.write("TIMESERIES EI_AEGI__BHZ, {0} samples, 16.2 sps, {1}, SLIST, INTEGER, Counts\n".format(packetSize, seedArray[0][0]))  # write header
        n = 0
        for line in seedArray:
            n += 1
            if (n == 1):
                text = repr(line[1]).rjust(10)
            else:
                text = text + repr(line[1]).rjust(12)
            if (n == 6):
                text = text + "  \n"
                asciiFile.write(text)
                n = 0
        text = text + "  \n"  # finish off any partial lines
        asciiFile.write(text)  # write the last line
        asciiFile.close()
        mseedFileName = seedArray[0][0].replace(":", "_")
        # there must be a symbolic link in the ringserver directory named "mseed" pointing to the location of the mseed files.
        command = 'ascii2mseed -r {0} -o /home/pi/ringserver/mseed/EI_AEGI__BHZ_{1}.mseed /home/pi/programs/Seismo/slist.ascii > /dev/null 2>&1 &'.format(packetSize,mseedFileName)
        seedArray = []
        writeFlag = 0
        os.system(command)

def signalProcessor(ts,tn, data):
    '''
    Pulse and step response swallower:
    Swallow data, by using previous datapoint, if difference between Capcount(n-1) and CapCount(n) is greater than X (TBD) and
    difference between CapCount(n+1) and CapCount(n-1) is less than X.
    If there is a step instead of a pulse, then take a rolling average of TBD points after the step edge and subtract that from
    CapCount data going forward. The amount of adjustment should slowly go to zero after a long period of time.
    '''
    global resultArray, seedArray
    if (len(resultArray) > 2):
        prevData = resultArray[-1][1]
        if ((prevData - data)^2 > 400^2 and prevData != resultArray[-2][1]):
            resultArray.append([tn, prevData])
            seedArray.append([ts.isoformat(),int(prevData)]) # raw data to seedlink server
            logger.info("{0} Swallowed Pulse {1}".format(datetime.now(),prevData - data))
        else:
            resultArray.append([tn, data])
            seedArray.append([ts.isoformat(),int(data)]) # raw data to seedlink server
    else:
        resultArray.append([tn, data])
        seedArray.append([ts.isoformat(),int(data)]) # raw data to seedlink server
    

def initial_plot_data():
    ''' return the data from the capCount array decimated by 8x '''
    global resultArray
    data = resultArray
    output=[]
    if (data):
        i = 0
        for row in data:
            if (i % 8 == 0):
                output.append( json.dumps({
                    'time' : row[0],
                    'capCount' : row[1]}))
            i += 1
        output2 = {"my_array" : output}
        #logger.info("my_array: {}".format(output2))
        return output2
    else:
        logger.error("No data obtained from resultArray for webpage.")
        timeNow = time.time()
        output.append( json.dumps({
                'time' : timeNow,
                'capCount' : 0}))
        output2 = {"my_array" : output}
        return output2

def update_plot_data(from_timestamp):
    ''' return the data from the capCount array decimated by 8x '''
    global resultArray, LPF1count
    #logger.info("LPF1count: {0:.2f}".format(LPF1count))
    data = resultArray
    output=[]

    if (data):
        lastIndex = findItem(data, float(from_timestamp))
        i = 0
        data = data[(lastIndex + 1):]
        for row in data:
            if (i % 8 == 0):
                output.append( json.dumps({
                    'time' : row[0],
                    'capCount' : row[1]}))
            i += 1
        output2 = {"my_array" : output}
        #logger.info("my_array: {}".format(output2))
        return output2
    else:
        logger.error("No update data obtained from resultArray for webpage.")
        timeNow = time.time()
        output.append( json.dumps({
                'time' : timeNow,
                'capCount' : 0}))
        output2 = {"my_array" : output}
        return output2

def findItem(theList, item):
    for i in range(1, len(theList)):
        part = theList[-i]
        if (item == part[0]):
            return -i
    return -1   # return the end of the list

def getTime():
   now = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
   return now

# run Bottle without any encryption
app = Bottle()

@app.route('/static/<filepath:path>')
def server_static(filepath):
    return static_file(filepath, root ='/home/pi/programs/Seismo/static_files')

# a simple json test main page
@app.route('/seismometer')
def seismoData():
    return template('/home/pi/programs/Seismo/seismometer.tpl')

@app.route('/plotData.json', method = 'POST')
def plotData():
    # get from/to data 
    from_ = request.POST.get('plotFrom','').strip()+"00:00"
    to_ = request.POST.get('plotTo','').strip()
    # returns flot data
    return initial_plot_data()

@app.route('/updateData.json', method = 'POST')
def updateData():
    from_ = request.POST.get('plotFrom','').strip()  # get the last timestamp
    #logger.info("Last Timestamp = {}".format(from_))
    # returns flot data
    return update_plot_data(from_)

logger.info("\n\t Seismometer is starting.\n")
# setup CapDACs and offset
cap2dig.writeCapDACA(0x80, debug=True)  # for 7.5pF cap = 0xb6
cap2dig.writeCapDACB(0x80, debug=True)  # for 10pf cap = 0xc7
#cap2dig.offsetCalibrate() # perform offset calibration
time.sleep(0.5)  # wait for offset calibration to complete

# GPIO4 is cap2dig conversion complete interrupt pin
GPIO.setmode(GPIO.BCM)
GPIO.setup(4, GPIO.IN)
GPIO.add_event_detect(4,GPIO.FALLING, callback=getData)

cap2dig.write_reg(0x0A, 0x21) # set for continuous conversion, 62ms


# remove mseed files more than 2 days old
def filecheck():
    while True:
        # check date and time
        utcTime = datetime.utcnow()
        thisHour = utcTime.hour
        if (thisHour == 0):
            twoDays = timedelta(days=2)
            fileDate = date.isoformat(date.today() - twoDays)
            # remove old files two days old
            try:
                logger.info("Removing old Files...{0}".format(fileDate))
                # there must be a symbolic link in the ringserver directory named "mseed" pointing to
                # the location of the mseed files 
                os.system('rm -f /home/pi/ringserver/mseed/*{0}* &'.format(fileDate))
            except:
                logger.error("Failed to remove old files.")
            time.sleep(60*60*2)  # sleep for 2 hours
        else:
            time.sleep(60*10)  # sleep for 10 minutes    

runfilecheck = multiprocessing.Process(target=filecheck)
runfilecheck.start()

run(app, host='192.168.1.12', port='8080',quiet=True)
runfilecheck.join()
